package club.clubk.ktag.konfigurator

val kTagByte = byteArrayOf(
    0x4B.toByte(), // K
    0x54.toByte(), // T
    0x61.toByte(), // a
    0x67.toByte()  // g
)

// Data classes for packet parsing
sealed class KTagPacket {
    abstract val type: Int
    abstract val typeName: String
    abstract val eventNumber: Int
}

data class InstigateGamePacket(
    override val type: Int,
    override val typeName: String = "Instigate Game",
    override val eventNumber: Int,
    val gameLength: Int,
    val countdownTime: Int
) : KTagPacket()

data class EventPacket(
    override val type: Int,
    override val typeName: String = "Event",
    override val eventNumber: Int,
    val targetAddress: String,
    val eventId: Int,
    val eventData: Int
) : KTagPacket()

data class TagPacket(
    override val type: Int,
    override val typeName: String = "Tag",
    override val eventNumber: Int,
    val txPower: Int,
    val protocol: Int,
    val teamId: Int,
    val playerId: Int,
    val damage: Int,
    val color: Int,
    val targetAddress: String
) : KTagPacket()

data class ConsolePacket(
    override val type: Int,
    override val typeName: String = "Console",
    override val eventNumber: Int,
    val consoleString: String
) : KTagPacket()

data class StatusPacket(
    override val type: Int,
    override val typeName: String = "Status",
    override val eventNumber: Int,
    val txPower: Int,
    val protocol: Int,
    val teamId: Int,
    val playerId: Int,
    val health: Int,
    val maxHealth: Int,
    val primaryColor: Int,
    val secondaryColor: Int,
    val systemKState: Int
) : KTagPacket()

data class ParametersPacket(
    override val type: Int,
    override val typeName: String = "Parameters",
    override val eventNumber: Int,
    val targetAddress: String,
    val subtype: Int,
    val key1: Int,
    val value1: Int,
    val key2: Int,
    val value2: Int
) : KTagPacket()

data class HelloPacket(
    override val type: Int,
    override val typeName: String = "Hello",
    override val eventNumber: Int,
    val majorVersion: Int,
    val minorVersion: Int,
    val deviceType: Int,
    val teamId: Int,
    val deviceName: String
) : KTagPacket()

fun byteArrayToKTagPacket(data: ByteArray): KTagPacket {
    val type = data[4].toInt()
    val eventNumber = data[5].toInt()

    return when (type) {
        0x01 -> InstigateGamePacket(
            type = type,
            eventNumber = eventNumber,
            gameLength = bytesToInt(data, 6, 4),
            countdownTime = bytesToInt(data, 10, 4)
        )

        0x02 -> EventPacket(
            type = type,
            eventNumber = eventNumber,
            targetAddress = bytesToMacAddress(data, 6),
            eventId = bytesToInt(data, 12, 4),
            eventData = bytesToInt(data, 16, 4)
        )

        0x03 -> TagPacket(
            type = type,
            eventNumber = eventNumber,
            txPower = data[6].toInt(),
            protocol = data[7].toInt(),
            teamId = data[8].toInt(),
            playerId = data[9].toInt(),
            damage = bytesToInt(data, 10, 2),
            color = bytesToInt(data, 12, 4),
            targetAddress = bytesToMacAddress(data, 16)
        )

        0x04 -> ConsolePacket(
            type = type,
            eventNumber = eventNumber,
            consoleString = data.slice(6..26).toByteArray().let { bytes ->
                // Find the position of the first null byte if any
                val nullPos = bytes.indexOfFirst { it == 0.toByte() }
                // If there's a null byte, take only up to that position, otherwise take all bytes
                if (nullPos >= 0) {
                    bytes.slice(0 until nullPos).toByteArray()
                } else {
                    bytes
                }.toString(Charsets.US_ASCII).trim()
            }
        )

        0x05 -> StatusPacket(
            type = type,
            eventNumber = eventNumber,
            txPower = data[6].toInt(),
            protocol = data[7].toInt(),
            teamId = data[8].toInt(),
            playerId = data[8].toInt(),
            health = bytesToInt(data, 10, 2),
            maxHealth = bytesToInt(data, 12, 2),
            primaryColor = bytesToInt(data, 14, 4),
            secondaryColor = bytesToInt(data, 18, 4),
            systemKState = data[22].toInt()
        )

        0x06 -> ParametersPacket(
            type = type,
            eventNumber = eventNumber,
            targetAddress = bytesToMacAddress(data, 6),
            subtype = data[12].toInt(),
            key1 = bytesToInt(data, 13, 2),
            value1 = bytesToInt(data, 15, 4),
            key2 = bytesToInt(data, 19, 2),
            value2 = bytesToInt(data, 21, 4)
        )

        0x07 -> HelloPacket(
            type = type,
            eventNumber = eventNumber,
            majorVersion = data[6].toInt(),
            minorVersion = data[7].toInt(),
            deviceType = bytesToInt(data, 8, 2),
            teamId = data[10].toInt(),
            deviceName = data.slice(11..26).toByteArray().let { bytes ->
                // Find the position of the first null byte if any
                val nullPos = bytes.indexOfFirst { it == 0.toByte() }
                // If there's a null byte, take only up to that position, otherwise take all bytes
                if (nullPos >= 0) {
                    bytes.slice(0 until nullPos).toByteArray()
                } else {
                    bytes
                }.toString(Charsets.US_ASCII).trim()
            }
        )

        else -> StatusPacket(
            type = type,
            typeName = "Unknown",
            eventNumber = eventNumber,
            txPower = 0,
            protocol = 0,
            teamId = 0,
            playerId = 0,
            health = 0,
            maxHealth = 0,
            primaryColor = 0,
            secondaryColor = 0,
            systemKState = 0
        )
    }
}

fun kTagPacketToByteArray(kTagPacket: KTagPacket): ByteArray {
    var payload: ByteArray = byteArrayOf(0x00.toByte())
    when (kTagPacket) {
        is ParametersPacket -> {
            val addressBytes = macAddressToBytes(kTagPacket.targetAddress)
            payload = byteArrayOf(
                0x06.toByte(), // Packet Type: Parameter
                (kTagPacket.eventNumber and 0xFF).toByte(), // Event counter

                *addressBytes,

                (kTagPacket.subtype and 0xFF).toByte(),

                (kTagPacket.key1 and 0xFF).toByte(),
                ((kTagPacket.key1 shr 8) and 0xFF).toByte(),

                (kTagPacket.value1 and 0xFF).toByte(),
                ((kTagPacket.value1 shr 8) and 0xFF).toByte(),
                ((kTagPacket.value1 shr 16) and 0xFF).toByte(),
                ((kTagPacket.value1 shr 24) and 0xFF).toByte(),

                (kTagPacket.key2 and 0xFF).toByte(),
                ((kTagPacket.key2 shr 8) and 0xFF).toByte(),

                (kTagPacket.value2 and 0xFF).toByte(),
                ((kTagPacket.value2 shr 8) and 0xFF).toByte(),
                ((kTagPacket.value2 shr 16) and 0xFF).toByte(),
                ((kTagPacket.value2 shr 24) and 0xFF).toByte()
            ) + ByteArray(2) { 0xFF.toByte() }
        }
        is ConsolePacket -> {
            // something
        }
        is EventPacket -> {
            val addressBytes = macAddressToBytes(kTagPacket.targetAddress)
            payload = byteArrayOf(
                0x02.toByte(), // Packet Type: Event
                (kTagPacket.eventNumber and 0xFF).toByte(), // Event counter

                *addressBytes,

                (kTagPacket.eventId and 0xFF).toByte(),
                ((kTagPacket.eventId shr 8) and 0xFF).toByte(),
                ((kTagPacket.eventId shr 16) and 0xFF).toByte(),
                ((kTagPacket.eventId shr 24) and 0xFF).toByte(), // Event ID

                (kTagPacket.eventData and 0xFF).toByte(),
                ((kTagPacket.eventData shr 8) and 0xFF).toByte(),
                ((kTagPacket.eventData shr 16) and 0xFF).toByte(),
                ((kTagPacket.eventData shr 24) and 0xFF).toByte() // Event data
            ) + ByteArray(7) { 0xFF.toByte() }
        }
        is HelloPacket -> {
            // something
        }
        is InstigateGamePacket -> {
            payload = byteArrayOf(
                0x01.toByte(), // Packet Type: Instigate Game
                (kTagPacket.eventNumber and 0xFF).toByte(), // Event counter

                (kTagPacket.gameLength and 0xFF).toByte(),
                ((kTagPacket.gameLength shr 8) and 0xFF).toByte(),
                ((kTagPacket.gameLength shr 16) and 0xFF).toByte(),
                ((kTagPacket.gameLength shr 24) and 0xFF).toByte(), // Game length

                (kTagPacket.countdownTime and 0xFF).toByte(),
                ((kTagPacket.countdownTime shr 8) and 0xFF).toByte(),
                ((kTagPacket.countdownTime shr 16) and 0xFF).toByte(),
                ((kTagPacket.countdownTime shr 24) and 0xFF).toByte() // Time until countdown
            ) + ByteArray(21) { 0xFF.toByte() }
        }
        is StatusPacket -> {
            // something
        }
        is TagPacket -> {
            // something
        }
    }
    return kTagByte + payload
}

// Helper function to convert bytes to integers
private fun bytesToInt(data: ByteArray, offset: Int, length: Int): Int {
    var result = 0
    for (i in 0 until length) {
        result = result or ((data[offset + i].toInt() and 0xFF) shl (8 * i))
    }
    return result
}

// Helper function to convert bytes to MAC address string
private fun bytesToMacAddress(data: ByteArray, offset: Int): String {
    return (0..5).joinToString(":") {
        String.format("%02X", data[offset + it])
    }
}

private fun macAddressToBytes(address: String): ByteArray {
    val cleanAddress = address.replace(":", "").replace("-", "")

    // Check if the address is valid (12 hex characters = 6 bytes)
    if (cleanAddress.length != 12) {
        throw IllegalArgumentException("Invalid BLE address length")
    }

    // Convert the hex string to bytes. Reverse the bytes to little-endian format.
    val bytes = ByteArray(6)
    var j = 5
    for (i in 0 until 6) {
        val hexByte = cleanAddress.substring(i * 2, i * 2 + 2)
        bytes[j] = hexByte.toInt(16).toByte()
        j--
    }

    return bytes
}

class InstigatePacketGenerator {
    private var eventCount = 0

    fun generatePacket(gameLength: Int, timeUntilCountdown: Int): InstigateGamePacket {
        val newInstigateGamePacket = InstigateGamePacket(
            type = 1, // instigate
            eventNumber = eventCount,
            gameLength = gameLength,
            countdownTime = timeUntilCountdown
        )
        eventCount++
        return newInstigateGamePacket
    }
}

class EventPacketGenerator {
    private var eventCount = 0

    fun generatePacket(targetAddress: String, eventId: Int = 0, eventData: Int = 0): EventPacket {
        val newEventPacket = EventPacket(
            type = 2, // event
            eventNumber = eventCount,
            targetAddress = targetAddress,
            eventId = eventId,
            eventData = eventData
        )
        eventCount++
        return newEventPacket
    }
}

class ParametersPacketGenerator {
    private var eventCount = 0

    fun generatePacket(targetAddress: String, subtype: Int,
                key1: Int = 0, value1: Int = 0,
                key2: Int = 0, value2: Int = 0): ParametersPacket {
        val newParametersPacket = ParametersPacket(
            type = 6, // parameters
            eventNumber = eventCount,
            targetAddress = targetAddress,
            subtype = subtype,
            key1 = key1, value1 = value1,
            key2 = key2, value2 = value2
        )
        eventCount++
        return newParametersPacket
    }
}