package club.clubk.ktag.konfigurator

import android.Manifest
import android.bluetooth.BluetoothManager
import android.content.Context
import android.content.pm.PackageManager
import android.os.Build
import android.os.Bundle
import android.util.Log
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.activity.result.contract.ActivityResultContracts
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.icons.filled.Check
import androidx.compose.material3.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.core.content.ContextCompat
// import androidx.lifecycle.viewmodel.compose.viewModel
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.res.painterResource
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.input.KeyboardType
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import club.clubk.ktag.konfigurator.ui.theme.KonfiguratorTheme
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch

private const val TAG = "Konfigurator"

class MainActivity : ComponentActivity() {
    private val requiredPermissions = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.S) {
        arrayOf(
            Manifest.permission.BLUETOOTH_ADVERTISE,
            Manifest.permission.BLUETOOTH_CONNECT,
            Manifest.permission.BLUETOOTH_SCAN,
            Manifest.permission.ACCESS_FINE_LOCATION,  // Required for BLE operations
            Manifest.permission.ACCESS_COARSE_LOCATION  // Required for BLE operations
        )
    } else {
        arrayOf(
            Manifest.permission.BLUETOOTH,
            Manifest.permission.BLUETOOTH_ADMIN,
            Manifest.permission.ACCESS_FINE_LOCATION,
            Manifest.permission.ACCESS_COARSE_LOCATION
        )
    }

    private val permissionLauncher = registerForActivityResult(
        ActivityResultContracts.RequestMultiplePermissions()
    ) { permissions ->
        val allGranted = permissions.entries.all { it.value }
        Log.d(TAG, "Permission results: ${permissions.map { "${it.key}: ${it.value}" }}")

        if (allGranted) {
            Log.i(TAG, "All permissions granted")
            mainScreen()
        } else {
            // Show which permissions were denied
            val deniedPermissions = permissions.filter { !it.value }.keys
            Log.w(TAG, "Some permissions denied: $deniedPermissions")
            setContent {
                MaterialTheme {
                    Surface(
                        modifier = Modifier.fillMaxSize(),
                        color = MaterialTheme.colorScheme.background
                    ) {
                        PermissionDeniedScreen(
                            deniedPermissions = deniedPermissions.toList()
                        ) { requestPermissions() }
                    }
                }
            }
        }
    }

    private fun checkAndRequestPermissions() {
        Log.d(TAG, "Checking permissions: ${requiredPermissions.joinToString()}")
        if (hasRequiredPermissions()) {
            Log.i(TAG, "All required permissions already granted")
            mainScreen()
        } else {
            Log.i(TAG, "Requesting permissions")
            requestPermissions()
        }
    }

    private fun hasRequiredPermissions(): Boolean {
        return requiredPermissions.all { permission ->
            ContextCompat.checkSelfPermission(this, permission) == PackageManager.PERMISSION_GRANTED
        }
    }

    private fun requestPermissions() {
        Log.d(TAG, "Launching permission request")
        permissionLauncher.launch(requiredPermissions)
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        checkAndRequestPermissions()
        // enableEdgeToEdge()
    }

    private fun mainScreen() {
        setContent {
            KonfiguratorTheme {
                StateMachine(StateMachineViewModel(applicationContext))
            }
        }
    }
}

@Composable
fun PermissionDeniedScreen(
    deniedPermissions: List<String>,
    onRequestPermissions: () -> Unit
) {
    Log.d(TAG, "Showing permission denied screen for permissions: $deniedPermissions")
    Column(
        modifier = Modifier
            .fillMaxSize()
            .padding(16.dp),
        horizontalAlignment = Alignment.CenterHorizontally,
        verticalArrangement = Arrangement.Center
    ) {
        Text(
            text = "The following permissions are required:",
            style = MaterialTheme.typography.bodyLarge,
            textAlign = TextAlign.Center,
            modifier = Modifier.padding(bottom = 8.dp)
        )
        deniedPermissions.forEach { permission ->
            Text(
                text = "• ${permission.split(".").last()}",
                style = MaterialTheme.typography.bodyMedium,
                textAlign = TextAlign.Center,
                modifier = Modifier.padding(vertical = 4.dp)
            )
        }
        Button(
            onClick = {
                Log.d(TAG, "Permission request button clicked")
                onRequestPermissions()
            },
            modifier = Modifier.padding(top = 16.dp)
        ) {
            Text("Grant Permissions")
        }
    }
}

@Composable
fun StateMachine(stateMachine: StateMachineViewModel) { // stateMachine: StateMachineViewModel = viewModel()
    val appState by stateMachine.appState.collectAsState()

    Surface(
        modifier = Modifier.fillMaxSize()
    ) {
        Column(
            modifier = Modifier
                .fillMaxSize()
                .padding(16.dp),
            horizontalAlignment = Alignment.CenterHorizontally,
            verticalArrangement = Arrangement.spacedBy(16.dp)
        ) {
            when (appState) {
                AppState.TitleScreen -> TitleScreen(stateMachine)
                AppState.GameSettings -> GameSettingsScreen(stateMachine)
                AppState.TeamSettings -> TeamSettingsScreen(stateMachine)
                AppState.PregameTimer -> PregameTimerScreen(stateMachine)
                AppState.Countdown -> CountdownScreen(stateMachine)
                AppState.GameTimer -> GameTimerScreen(stateMachine)
                AppState.GameOver -> GameOverScreen(stateMachine)
                AppState.WrapUp -> Text("Wrap Up")
            }
        }
    }
}

// ------------
// TITLE SCREEN
//

@Composable
fun BigLogo() {
    Icon(
        painter = painterResource(id = R.drawable.konfigurator),
        contentDescription = "Konfigurator Logo",
        modifier = Modifier.size(192.dp),
        tint = Color.Unspecified
    )
}

@Composable
fun TitleScreen(stateMachine: StateMachineViewModel) {
    Column(
        modifier = Modifier.fillMaxWidth(),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        Text("Konfigurator", style = MaterialTheme.typography.headlineMedium)
        BigLogo()
    }
    Spacer(modifier = Modifier.height(16.dp))
    Button(
        onClick = { stateMachine.nextState() },
        modifier = Modifier.fillMaxWidth()
    ) {
        Text("New Game")
    }
}

// -------------
// GAME SETTINGS
// -------------

@Composable
fun GameSettingsScreen(stateMachine: StateMachineViewModel) {
    val currentGameConfig by stateMachine.currentGameConfig.collectAsState()

    Column(
        modifier = Modifier.fillMaxSize(),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        Text(
            text = "Game Settings",
            style = MaterialTheme.typography.headlineMedium,
            modifier = Modifier.padding(16.dp)
        )
        GameConfigEditor(oldGameConfig = currentGameConfig,
            onSave = {
                stateMachine.updateGameConfig(it)
                stateMachine.nextState()
            })
    }
}

@Composable
fun GameConfigEditor(oldGameConfig: GameConfig,
                     onSave: (GameConfig) -> Unit,
                     modifier: Modifier = Modifier) {
    key(oldGameConfig) {
        var name by rememberSaveable(oldGameConfig.name) { mutableStateOf(oldGameConfig.name) }
        var gameLength by rememberSaveable(oldGameConfig.gameLength) { mutableStateOf(oldGameConfig.gameLength.toString()) }
        var pregameLength by rememberSaveable(oldGameConfig.pregameLength) { mutableStateOf(oldGameConfig.pregameLength.toString()) }
        var numRounds by rememberSaveable(oldGameConfig.numRounds) { mutableStateOf(oldGameConfig.numRounds.toString()) }
        var maxHealth by rememberSaveable(oldGameConfig.maxHealth) { mutableStateOf(oldGameConfig.maxHealth.toString()) }
        var numBombs by rememberSaveable(oldGameConfig.numBombs) { mutableStateOf(oldGameConfig.numBombs.toString()) }

        // For tracking validation errors
        var gameLengthError by rememberSaveable { mutableStateOf(false) }
        var pregameLengthError by rememberSaveable { mutableStateOf(false) }
        var numRoundsError by rememberSaveable { mutableStateOf(false) }
        var maxHealthError by rememberSaveable { mutableStateOf(false) }
        var numBombsError by rememberSaveable { mutableStateOf(false) }

        // Function to validate the form
        fun validateForm(): Boolean {
            gameLengthError = !isValidInteger(gameLength)
            pregameLengthError = !isValidInteger(pregameLength)
            numRoundsError = !isValidInteger(numRounds)
            maxHealthError = !isValidInteger(maxHealth)
            numBombsError = !isValidInteger(numBombs)

            return !gameLengthError && !pregameLengthError && !numRoundsError && !maxHealthError && !numBombsError
        }

        // Function to update the data class when values change
        fun saveGameConfig() {
            // Only update if all integer fields are valid
            if (validateForm()) {
                val newGameConfig = GameConfig(
                    name = name,
                    gameLength = gameLength.toIntOrNull() ?: oldGameConfig.gameLength,
                    pregameLength = pregameLength.toIntOrNull() ?: oldGameConfig.pregameLength,
                    numRounds = numRounds.toIntOrNull() ?: oldGameConfig.numRounds,
                    maxHealth = maxHealth.toIntOrNull() ?: oldGameConfig.maxHealth,
                    numBombs = numBombs.toIntOrNull() ?: oldGameConfig.numBombs
                )
                onSave(newGameConfig)
            }
        }

        Column(
            modifier = modifier
                .fillMaxWidth()
                .padding(16.dp),
            verticalArrangement = Arrangement.spacedBy(8.dp)
        ) {
            // String value field
            OutlinedTextField(
                value = name,
                onValueChange = {
                    name = it
                },
                label = { Text("Name") },
                modifier = Modifier.fillMaxWidth()
            )

            // Integer value fields
            IntegerTextField(
                value = gameLength,
                onValueChange = {
                    gameLength = it
                    // gameLengthError = !isValidInteger(it)
                },
                label = "Game Length",
                isError = gameLengthError
            )

            IntegerTextField(
                value = pregameLength,
                onValueChange = {
                    pregameLength = it
                    // pregameLengthError = !isValidInteger(it)
                },
                label = "Pregame Length",
                isError = pregameLengthError
            )

            IntegerTextField(
                value = numRounds,
                onValueChange = {
                    numRounds = it
                    // numRoundsError = !isValidInteger(it)
                },
                label = "Number Of Rounds",
                isError = numRoundsError
            )

            IntegerTextField(
                value = maxHealth,
                onValueChange = {
                    maxHealth = it
                    // maxHealthError = !isValidInteger(it)
                },
                label = "Max Health",
                isError = maxHealthError
            )

            IntegerTextField(
                value = numBombs,
                onValueChange = {
                    numBombs = it
                    // numBombsError = !isValidInteger(it)
                },
                label = "Number Of Bombs",
                isError = numBombsError
            )
        }

        Spacer(modifier = Modifier.height(8.dp))
        Button( // Save button
            onClick = { saveGameConfig() },
            modifier = Modifier.fillMaxWidth()
        ) {
            Text("Confirm")
        }
    }
}

// Helper function to validate integer input
private fun isValidInteger(input: String): Boolean {
    return input.toIntOrNull() != null || input.isEmpty()
}

// Helper composable for integer input fields
@Composable
private fun IntegerTextField(
    value: String,
    onValueChange: (String) -> Unit,
    label: String,
    isError: Boolean
) {
    OutlinedTextField(
        value = value,
        onValueChange = onValueChange,
        label = { Text(label) },
        keyboardOptions = KeyboardOptions(keyboardType = KeyboardType.Number),
        isError = isError,
        supportingText = if (isError) {
            { Text("Please enter a valid integer") }
        } else null,
        modifier = Modifier.fillMaxWidth()
    )
}

// -------------
// TEAM SETTINGS
// -------------

@Composable
fun TeamSettingsScreen(stateMachine: StateMachineViewModel) {
    Log.d("MAINACTIVITY", "recomposing team setting screen")
    val coroutineScope = rememberCoroutineScope()
    val devices by stateMachine.devices.collectAsState()
    val allDevicesReady by stateMachine.allDevicesReady.collectAsState()
    Scaffold(
        // Bottom bar that stays fixed at the bottom
        bottomBar = {
            BottomAppBar(
                containerColor = MaterialTheme.colorScheme.surface,
                tonalElevation = 8.dp
            ) {
                Box(
                    modifier = Modifier.fillMaxWidth(),
                    contentAlignment = Alignment.Center
                ) {
                    if (allDevicesReady) {
                        Button(
                            onClick = { stateMachine.nextState() },
                            modifier = Modifier.fillMaxWidth()
                        ) {
                            Text("Start Game")
                        }
                    } else {
                        SideBySideButtons(
                            "Apply",
                            "Ready",
                            onLeftButtonClick = { coroutineScope.launch { stateMachine.configureDevices() } },
                            onRightButtonClick = { stateMachine.readyAllDevices() }
                        )
                    }
                }
            }
        }
    ) { paddingValues ->
        DeviceList(stateMachine, devices, Modifier.fillMaxSize().padding(paddingValues))
    }
}

@Composable
fun DeviceList(stateMachine: StateMachineViewModel, devices: List<Device>, modifier: Modifier) {
    LazyColumn(modifier = modifier) {
        items(devices) { device ->
            DeviceCard(stateMachine, device)
        }
    }
}

@Composable
fun SideBySideButtons(
    leftButtonText: String,
    rightButtonText: String,
    onLeftButtonClick: () -> Unit,
    onRightButtonClick: () -> Unit,
    modifier: Modifier = Modifier
) {
    Row(
        modifier = modifier
            .fillMaxWidth()
            .padding(16.dp),
        horizontalArrangement = Arrangement.spacedBy(8.dp) // Add space between buttons
    ) {
        Button(
            onClick = onLeftButtonClick,
            modifier = Modifier.weight(1f) // Takes up half of the available width
        ) {
            Text(text = leftButtonText)
        }

        Button(
            onClick = onRightButtonClick,
            modifier = Modifier.weight(1f) // Takes up half of the available width
        ) {
            Text(text = rightButtonText)
        }
    }
}

@Composable
fun DeviceCard(stateMachine: StateMachineViewModel, device: Device) {
    val backgroundColor = when (device.team) {
        0 -> Color.Magenta
        1 -> Color.Red
        2 -> Color.Blue
        else -> MaterialTheme.colorScheme.surface // Default color
    }
    Card(modifier = Modifier
        .fillMaxWidth()
        .padding(vertical = 4.dp),
        onClick = { stateMachine.cycleDeviceTeam(device) },
        colors = CardDefaults.cardColors(
            containerColor = backgroundColor
        )) {
        Text(device.name + " | " + device.deviceTypeName())
        Text(if(device.deviceState == DeviceState.Ready) {"READY"} else {"NOT READY"})
    }
}

// -------------
// PREGAME TIMER
// -------------

fun formatTimeToMinutesSeconds(remainingSeconds: Int): String {
    val minutes = remainingSeconds / 60
    val seconds = remainingSeconds % 60

    // Format to ensure seconds always shows with two digits (e.g., "01" instead of "1")
    return "$minutes:${seconds.toString().padStart(2, '0')}"
}

@Composable
fun CountdownTimer(timeInMillis: Long, onCountdownComplete: () -> Unit) {
    // Store remaining time as state
    var remainingTime by remember { mutableLongStateOf(timeInMillis) }
    val initialTime = remember { timeInMillis }

    // Use LaunchedEffect to create a coroutine that runs the countdown
    LaunchedEffect(key1 = timeInMillis) {
        // Reset the timer if the input time changes
        remainingTime = timeInMillis

        // Continue counting down until we reach zero
        while (remainingTime > 0) {
            delay(1000) // Update every second
            remainingTime -= 1000
            Log.d("CountdownTimer", "Remaining time: ${remainingTime / 1000} seconds")
        }

        // Log when countdown reaches zero
        Log.d("CountdownTimer", "Countdown finished!")
        onCountdownComplete()
    }

    val remainingSeconds = (remainingTime / 1000).toInt()
    val progress = if (initialTime > 0) remainingTime.toFloat() / initialTime.toFloat() else 0f
    val formattedTime = formatTimeToMinutesSeconds(remainingSeconds)

    // UI to display the countdown
    Column(
        modifier = Modifier
            .fillMaxWidth()
            .padding(16.dp),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        // Display remaining time in seconds
        Text(
            text = formattedTime,
            fontSize = 36.sp,
            fontWeight = FontWeight.Bold,
            modifier = Modifier.padding(bottom = 8.dp)
        )

        // Display progress bar
        LinearProgressIndicator(
            progress = { progress },
            modifier = Modifier.fillMaxWidth(),
        )
    }
}

@Composable
fun PregameTimerScreen(stateMachine: StateMachineViewModel) {
    val currentGameConfig by stateMachine.currentGameConfig.collectAsState()
    Column(
        modifier = Modifier.fillMaxWidth(),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        Text(
            text = "Game Starts In",
            style = MaterialTheme.typography.headlineMedium
        )
        CountdownTimer(currentGameConfig.pregameLength.toLong(), onCountdownComplete = {
            stateMachine.nextState()
        })
    }
}

// ---------
// COUNTDOWN
// ---------

@Composable
fun CountdownScreen(stateMachine: StateMachineViewModel) {
    Column(
        modifier = Modifier.fillMaxWidth(),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        CountdownTimer(6000.toLong(), onCountdownComplete = {
            stateMachine.nextState()
        })
    }
}

// ----------
// GAME TIMER
// ----------

@Composable
fun GameTimerScreen(stateMachine: StateMachineViewModel) {
    val currentGameConfig by stateMachine.currentGameConfig.collectAsState()
    Column(
        modifier = Modifier.fillMaxWidth(),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        Text(
            text = "Game In Progress",
            style = MaterialTheme.typography.headlineMedium
        )

        CountdownTimer(currentGameConfig.gameLength.toLong(), onCountdownComplete = {
            stateMachine.nextState()
        })

        Button(
            onClick = { stateMachine.nextState() },
            modifier = Modifier.fillMaxWidth()
        ) {
            Text("End Game")
        }
    }
}

// ---------
// GAME OVER
// ---------

@Composable
fun GameOverScreen(stateMachine: StateMachineViewModel) {
    Column(
        modifier = Modifier.fillMaxWidth(),
        horizontalAlignment = Alignment.CenterHorizontally
    ) {
        Text(
            text = "Game Over",
            style = MaterialTheme.typography.headlineMedium
        )

        Spacer(modifier = Modifier.height(24.dp))

        Icon(
            imageVector = androidx.compose.material.icons.Icons.Default.Check,
            contentDescription = null,
            modifier = Modifier.size(100.dp),
            tint = MaterialTheme.colorScheme.primary
        )

        Spacer(modifier = Modifier.height(16.dp))
        Button(
            onClick = { stateMachine.nextState() },
            modifier = Modifier.fillMaxWidth()
        ) {
            Text("Play Again")
        }
    }
}