package club.clubk.ktag.konfigurator

import android.util.Log
import kotlinx.coroutines.*
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock

// Assuming ParametersPacket and BleManager are defined as in your existing code

class MultiDeviceConfigurator(private val bleManager: BleManager) {

    // Mutex to ensure only one device configuration (advertise + scan for ack) happens at a time
    private val bleOperationMutex = Mutex()

    /**
     * Configures a list of devices sequentially in terms of BLE operations,
     * but the overall process can be managed as part of a larger concurrent workflow.
     *
     * @param deviceConfigs List of ParametersPacket for each device to configure.
     * @param onDeviceConfigured A callback invoked for each device with its address and success status.
     * @param onProgressUpdate A callback to report overall progress (e.g., "Configuring device X of Y").
     * @return A map of device addresses to their configuration success status.
     */
    suspend fun configureMultipleDevices(
        deviceConfigs: List<ParametersPacket>,
        onDeviceConfigured: suspend (deviceAddress: String, success: Boolean) -> Unit,
        onProgressUpdate: suspend (configuredCount: Int, totalCount: Int) -> Unit
    ): Map<String, Boolean> {
        val results = mutableMapOf<String, Boolean>()
        var configuredCount = 0
        val totalCount = deviceConfigs.size

        onProgressUpdate(configuredCount, totalCount)

        for (configPacket in deviceConfigs) {
            // Ensure only one BLE configuration operation (advertising + specific scan)
            // happens at a time across all calls that might use this shared BleManager.
            // If configureDevice itself manages a Mutex for its specific startAdvertising/startScanning,
            // this outer Mutex might be redundant for that part, but it's good for clarity
            // if other BLE operations could interfere.
            // For now, let's assume configureDevice is a self-contained operation.
            // The critical part is that `bleManager.configureDevice` itself does a full
            // advertise -> scan -> stop cycle.

            var success = false
            try {
                // The bleOperationMutex ensures that if multiple coroutines call
                // configureMultipleDevices, or if other parts of your app try to use
                // bleManager for advertising/scanning, they don't interfere with the
                // current device's configuration sequence.
                bleOperationMutex.withLock {
                    Log.i(
                        "MultiDeviceConfigurator",
                        "Attempting to configure ${configPacket.targetAddress}"
                    )
                    // The existing configureDevice already has its own timeout.
                    success = bleManager.configureDevice(configPacket)
                }
            } catch (e: Exception) {
                // Catch any unexpected exceptions during the configuration of a single device
                Log.e(
                    "MultiDeviceConfigurator",
                    "Exception configuring ${configPacket.targetAddress}",
                    e
                )
                success = false
            }

            results[configPacket.targetAddress] = success
            configuredCount++
            onDeviceConfigured(configPacket.targetAddress, success)
            onProgressUpdate(configuredCount, totalCount)

            if (success) {
                Log.i(
                    "MultiDeviceConfigurator",
                    "Successfully configured ${configPacket.targetAddress}"
                )
            } else {
                Log.w(
                    "MultiDeviceConfigurator",
                    "Failed to configure ${configPacket.targetAddress}"
                )
            }

            // Optional: Add a small delay between configuring devices if peripherals need time
            // delay(100)
        }
        return results
    }
}

// --- Example Usage (e.g., in a ViewModel or a Service) ---
suspend fun exampleUsage(bleManager: BleManager, allDeviceConfigs: List<ParametersPacket>) {
    val configurator = MultiDeviceConfigurator(bleManager)

    Log.d("App", "Starting configuration for ${allDeviceConfigs.size} devices...")

    // This whole block can be launched in a specific coroutine scope (e.g., viewModelScope)
    // The configureMultipleDevices function itself is sequential for BLE ops,
    // but the call to it can be non-blocking for the UI.
    val configurationResults = configurator.configureMultipleDevices(
        deviceConfigs = allDeviceConfigs,
        onDeviceConfigured = { deviceAddress, success ->
            Log.d("App", "Device $deviceAddress configuration result: $success")
            // Update UI for this specific device
        },
        onProgressUpdate = { configured, total ->
            Log.d("App", "Progress: $configured / $total devices configured.")
            // Update overall progress UI
        }
    )

    Log.d("App", "All device configurations attempted. Results: $configurationResults")
    // Process overall results
}