package club.clubk.ktag.konfigurator

import android.annotation.SuppressLint
import android.bluetooth.le.AdvertiseCallback
import android.bluetooth.le.AdvertiseSettings
import android.bluetooth.le.ScanCallback
import android.bluetooth.le.ScanResult
import android.content.Context
import android.util.Log
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.asStateFlow
import kotlinx.coroutines.flow.update

private const val TAG_BLE_SCAN = "BLE Scanner"
private const val TAG_BLE_AD = "BLE Advertiser"
private const val TAG_STATE_MACHINE = "StateMachineVM"

class StateMachineViewModel(context: Context) : ViewModel() {

    private val _appState = MutableStateFlow<AppState>(AppState.TitleScreen)
    val appState: StateFlow<AppState> = _appState.asStateFlow()

    private val _devices = MutableStateFlow<List<Device>>(emptyList())
    val devices: StateFlow<List<Device>> = _devices.asStateFlow()

    private val _allDevicesReady = MutableStateFlow<Boolean>(false)
    val allDevicesReady = _allDevicesReady.asStateFlow()

    private val _currentGameConfig = MutableStateFlow(GameConfig())
    val currentGameConfig = _currentGameConfig.asStateFlow()

    // State for configuration progress
    private val _configurationProgress = MutableStateFlow<ConfigurationProgress?>(null)
    val configurationProgress: StateFlow<ConfigurationProgress?> = _configurationProgress.asStateFlow()

    // State for individual device configuration status
    private val _deviceConfigStatus = MutableStateFlow<Map<String, Boolean>>(emptyMap())
    val deviceConfigStatus: StateFlow<Map<String, Boolean>> = _deviceConfigStatus.asStateFlow()

    private val bleManager = BleManager.getInstance(context)
    private val instigatePacketGenerator = InstigatePacketGenerator()
    private val eventPacketGenerator = EventPacketGenerator()
    private val parameterPacketGenerator = ParametersPacketGenerator()
    private val multiDeviceConfigurator = MultiDeviceConfigurator(bleManager)

    private val soundManager = SoundManager(context)

    override fun onCleared() {
        soundManager.release()
        super.onCleared()
    }

    @SuppressLint("MissingPermission")
    private val scanCallbackDefault = object : ScanCallback() {
        override fun onScanResult(callbackType: Int, result: ScanResult) {
            result.scanRecord?.manufacturerSpecificData?.get(0xFFFF)?.let { data ->
                if (data.size >= 4 &&
                    data[0] == 0x4B.toByte() && // K
                    data[1] == 0x54.toByte() && // T
                    data[2] == 0x61.toByte() && // a
                    data[3] == 0x67.toByte()    // g
                ) {
                    val packet = byteArrayToKTagPacket(data)
                    val scannedDevice = Device(address = result.device.address)
                    when (packet) {
                        is HelloPacket -> {
                            // Log.d(TAG_BLE_SCAN, "HelloPacket scanned")
                            scannedDevice.name = packet.deviceName
                            scannedDevice.deviceType = packet.deviceType
                            scannedDevice.team = packet.teamId
                            scannedDevice.deviceState = DeviceState.Configurable
                            addOrRefreshDevice(scannedDevice)
                        }
                        is ConsolePacket -> {
                            // Log.d(TAG_BLE_SCAN, "ConsolePacket scanned")
                        }
                        is EventPacket -> {
                            // Log.d(TAG_BLE_SCAN, "EventPacket scanned")
                        }
                        is InstigateGamePacket -> {
                            // Log.d(TAG_BLE_SCAN, "InstigateGamePacket scanned")
                        }
                        is ParametersPacket -> {
                            // Log.d(TAG_BLE_SCAN, "ParametersPacket scanned")
                        }
                        is StatusPacket -> {
                            when (packet.systemKState) {
                                3 -> scannedDevice.deviceState = DeviceState.Ready
                                7 -> scannedDevice.deviceState = DeviceState.Playing
                                9 -> scannedDevice.deviceState = DeviceState.WrapUp
                            }
                            addOrRefreshDevice(scannedDevice)
                        }
                        is TagPacket -> {
                            // Log.d(TAG_BLE_SCAN, "TagPacket scanned")
                        }
                    }
                }
            }
        }

        override fun onScanFailed(errorCode: Int) {
            val errorMessage = when (errorCode) {
                SCAN_FAILED_ALREADY_STARTED -> "Already started"
                SCAN_FAILED_APPLICATION_REGISTRATION_FAILED -> "App registration failed"
                SCAN_FAILED_FEATURE_UNSUPPORTED -> "Feature unsupported"
                SCAN_FAILED_INTERNAL_ERROR -> "Internal error"
                else -> "Unknown error $errorCode"
            }
            Log.e(TAG_BLE_SCAN, "BLE Scan failed: $errorMessage")
        }
    }

    private val advertisingCallbackDefault = object : AdvertiseCallback() {
        override fun onStartSuccess(settingsInEffect: AdvertiseSettings) {
            super.onStartSuccess(settingsInEffect)
            Log.i(
                TAG_BLE_AD, "Advertisement started successfully with settings: " +
                        "Mode=${settingsInEffect.mode}, TxPower=${settingsInEffect.txPowerLevel}"
            )
        }

        override fun onStartFailure(errorCode: Int) {
            super.onStartFailure(errorCode)
            val errorMessage = when (errorCode) {
                ADVERTISE_FAILED_ALREADY_STARTED -> "Already started"
                ADVERTISE_FAILED_DATA_TOO_LARGE -> "Data too large"
                ADVERTISE_FAILED_TOO_MANY_ADVERTISERS -> "Too many advertisers"
                ADVERTISE_FAILED_INTERNAL_ERROR -> "Internal error"
                ADVERTISE_FAILED_FEATURE_UNSUPPORTED -> "Feature unsupported"
                else -> "Unknown error $errorCode"
            }
            Log.e(TAG_BLE_AD, "Failed to start advertising: $errorMessage")
        }
    }

    private fun addOrRefreshDevice(newDevice: Device) {
        val currentDevices = _devices.value.toMutableList()
        val index = currentDevices.indexOfFirst { it.address == newDevice.address }
        if (index == -1) {
            currentDevices += newDevice
        }
        else {
            var oldDevice = currentDevices[index]
            newDevice.name = oldDevice.name
            newDevice.deviceType = oldDevice.deviceType ?: newDevice.deviceType
            newDevice.team = oldDevice.team ?: newDevice.team
            newDevice.playerID = oldDevice.playerID ?: newDevice.playerID
            newDevice.deviceState = newDevice.deviceState ?: oldDevice.deviceState
            currentDevices[index] = newDevice
        }
        _devices.value = currentDevices
        Log.d("AddRefresh", _devices.value.toString())
        _allDevicesReady.value = allDevicesReady()
        Log.d("AddRefresh", _allDevicesReady.value.toString())
    }

    private fun updateDevice(newDevice: Device) {
        val currentDevices = _devices.value.toMutableList()
        val index = currentDevices.indexOfFirst { it.address == newDevice.address }
        if (index == -1) { return }
        var oldDevice = currentDevices[index]
        newDevice.deviceType = newDevice.deviceType ?: oldDevice.deviceType
        newDevice.team = newDevice.team ?: oldDevice.team
        newDevice.playerID = newDevice.playerID ?: oldDevice.playerID
        currentDevices[index] = newDevice
        _devices.value = currentDevices
        _allDevicesReady.value = allDevicesReady()
    }

    fun updateDeviceTeam(deviceAddress: String, newTeam: Int) {
        _devices.update { currentList ->
            currentList.map { device ->
                if (device.address == deviceAddress) {
                    device.copy(team = newTeam) // Creates a new Device instance
                } else {
                    device
                }
            }
        }
        _allDevicesReady.value = allDevicesReady()
    }

    fun cycleDeviceTeam(device: Device) {
        Log.d("STATEMACHINE", "cycling device team")
        var newTeam = device.team ?: -1
        newTeam++
        if (newTeam > 2) {
            newTeam = 0
        }
        updateDeviceTeam(device.address, newTeam)
    }

    private fun allDevicesReady(): Boolean {
        if (_devices.value.isEmpty()) { return false }
        for (device in _devices.value) {
            if (device.deviceState != DeviceState.Ready) {
                return false
            }
        }
        return true
    }

    fun readyAllDevices() {
        val eventPacket = eventPacketGenerator.generatePacket(targetAddress = "FF:FF:FF:FF:FF:FF", eventId = 2)
        bleManager.startAdvertising(kTagPacketToByteArray(eventPacket), advertisingCallbackDefault)
        bleManager.startScanning(scanCallbackDefault)
    }

    fun updateGameConfig(newGameConfig: GameConfig) {
        _currentGameConfig.value = newGameConfig
    }

    /**
     * Initiates the configuration of all devices using the MultiDeviceConfigurator.
     */
    fun configureDevices() {
        val currentDevicesToConfigure = _devices.value.filter {
            // Refine this filter as needed. For example:
            // it.deviceState == DeviceState.Configurable || it.needsReconfiguration
            true // For now, assuming all devices in the list should be configured
        }
        val gameCfg = _currentGameConfig.value

        if (currentDevicesToConfigure.isEmpty()) {
            Log.i(TAG_STATE_MACHINE, "No devices eligible for configuration.")
            _configurationProgress.value = ConfigurationProgress(0, 0, true)
            return
        }

        // 1. Prepare the list of ParametersPacket (this part is similar to what you had,
        // but now it collects all packets for all devices first)
        val deviceConfigPackets = currentDevicesToConfigure.mapNotNull { device ->
            val teamId = device.team ?: 0 // Default to team 0 if not set

            // Create your ParametersPacket.
            // If a device can have multiple configuration packets, generate them all here
            // and return a List<ParametersPacket> from this lambda, then use flatMapNotNull.
            // For now, assuming one packet per device for simplicity.
            parameterPacketGenerator.generatePacket(
                targetAddress = device.address,
                subtype = 2, // Request Parameter Change
                key1 = 1, value1 = teamId,                 // Key 1 is Team ID
                key2 = 4, value2 = gameCfg.maxHealth       // Key 2 is Max Health
            )
            // If a device for some reason can't be configured (e.g. missing address),
            // you could return null here, and mapNotNull will filter it out.
        }

        if (deviceConfigPackets.isEmpty() && currentDevicesToConfigure.isNotEmpty()) {
            Log.w(TAG_STATE_MACHINE, "Could not generate config packets for any selected device.")
            _configurationProgress.value = ConfigurationProgress(0, currentDevicesToConfigure.size, true)
            return
        }

        Log.i(TAG_STATE_MACHINE, "Starting configuration for ${deviceConfigPackets.size} devices.")
        _deviceConfigStatus.value = emptyMap() // Reset status for the new session
        _configurationProgress.value = ConfigurationProgress(0, deviceConfigPackets.size, false)

        viewModelScope.launch {
            // Optional: Stop general scanning if it might interfere with the
            // specific scan/advertise sequence within bleManager.configureDevice()
            // Log.d(TAG_STATE_MACHINE, "Stopping general BLE scan for configuration.")
            // bleManager.stopScanning()

            // 2. THIS IS THE REPLACEMENT: Call your MultiDeviceConfigurator
            val results = multiDeviceConfigurator.configureMultipleDevices(
                deviceConfigs = deviceConfigPackets,
                onDeviceConfigured = { deviceAddress, success ->
                    // This callback is invoked by your MultiDeviceConfigurator for each device
                    Log.i(TAG_STATE_MACHINE, "Device $deviceAddress configured (from callback): $success")
                    _deviceConfigStatus.update { currentStatus ->
                        currentStatus + (deviceAddress to success)
                    }
                    // You could also update the Device object in _devices if needed:
                    // _devices.update { list ->
                    // list.map { if (it.address == deviceAddress) it.copy(isConfigured = success) else it }
                    // }
                },
                onProgressUpdate = { configuredCount, totalCount ->
                    // This callback is invoked by your MultiDeviceConfigurator on progress
                    Log.i(TAG_STATE_MACHINE, "Configuration progress (from callback): $configuredCount/$totalCount")
                    _configurationProgress.value = ConfigurationProgress(configuredCount, totalCount, configuredCount == totalCount)
                }
            )

            // 3. Process overall results (optional, as callbacks already update state)
            Log.i(TAG_STATE_MACHINE, "All device configurations attempted by MultiDeviceConfigurator. Final Results map: $results")
            val allSucceeded = results.values.all { it }
            if (allSucceeded) {
                Log.i(TAG_STATE_MACHINE, "All devices configured successfully (overall check).")
            } else {
                Log.w(TAG_STATE_MACHINE, "One or more devices failed to configure (overall check).")
            }

            // Optional: Restart general scanning if it was stopped earlier
            // Log.d(TAG_STATE_MACHINE, "Restarting general BLE scan after configuration.")
            // bleManager.startScanning(scanCallbackDefault)
        }
    }

    fun nextState() {
        soundManager.playDoubleChirp()
        when (_appState.value) {
            is AppState.TitleScreen -> {
                _appState.value = AppState.GameSettings
            }
            is AppState.GameSettings -> {
                _appState.value = AppState.TeamSettings
                bleManager.startScanning(scanCallbackDefault)
            }
            is AppState.TeamSettings -> {
                _appState.value = AppState.PregameTimer
                val instigatePacket = instigatePacketGenerator.generatePacket(
                    _currentGameConfig.value.gameLength,
                    _currentGameConfig.value.pregameLength
                )
                bleManager.stopScanning()
                bleManager.startAdvertising(kTagPacketToByteArray(instigatePacket), advertisingCallbackDefault)
            }
            is AppState.PregameTimer -> {
                _appState.value = AppState.Countdown
                bleManager.stopAdvertising()
            }
            is AppState.Countdown -> {
                _appState.value = AppState.GameTimer
            }
            is AppState.GameTimer -> {
                _appState.value = AppState.GameOver
                val gameOverPacket = eventPacketGenerator.generatePacket(targetAddress = "FF:FF:FF:FF:FF:FF", eventId = 4)
                bleManager.startAdvertising(kTagPacketToByteArray(gameOverPacket), advertisingCallbackDefault)
            }
            AppState.GameOver -> {
                _appState.value = AppState.GameSettings
                val wrapupCompletePacket = eventPacketGenerator.generatePacket(targetAddress = "FF:FF:FF:FF:FF:FF", eventId = 3, eventData = 2)
                bleManager.startAdvertising(kTagPacketToByteArray(wrapupCompletePacket), advertisingCallbackDefault)
            }
            AppState.WrapUp -> {
            _appState.value = AppState.GameSettings
            }
        }
    }
}

